@@ -14,7 +14,7 @@ __maintainer__ = "Giridhar Pemmasani (pgiri@yahoo.com)"
 __license__ = "MIT"
 __url__ = "http://dispy.sourceforge.net"
 __status__ = "Production"
-__version__ = "3.15"
+__version__ = "3.16"
 
 __all__ = ['logger', 'DispyJob', 'JobCluster', 'SharedJobCluster']
 
@@ -42,7 +42,7 @@ import queue
 import numbers
 
 import asyncoro
-from asyncoro import Coro, AsynCoro, AsynCoroSocket, MetaSingleton, serialize, unserialize
+from asyncoro import Coro, AsynCoro, AsyncSocket, MetaSingleton, serialize, unserialize
 
 _dispy_version = __version__
 
@@ -273,7 +273,7 @@ class _Node(object):
         # generator
         # logger.debug('Sending to %s:%s', self.ip_addr, self.port)
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, keyfile=self.keyfile, certfile=self.certfile)
+        sock = AsyncSocket(sock, keyfile=self.keyfile, certfile=self.certfile)
         sock.settimeout(5)
         try:
             yield sock.connect((self.ip_addr, self.port))
@@ -296,7 +296,7 @@ class _Node(object):
         # generator
         # logger.debug('XferFile: %s to %s:%s', xf.name, self.ip_addr, self.port)
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, keyfile=self.keyfile, certfile=self.certfile)
+        sock = AsyncSocket(sock, keyfile=self.keyfile, certfile=self.certfile)
         sock.settimeout(5)
         try:
             yield sock.connect((self.ip_addr, self.port))
@@ -367,17 +367,14 @@ class _DispyJob_(object):
                 if dep in depend_ids:
                     continue
                 try:
+                    sbuf = os.stat(dep)
+                    # TODO: Check/limit size
                     fd = open(dep, 'rb')
+                    data = fd.read()
                     fd.close()
                 except:
                     logger.warning('File "%s" is not valid' % dep)
                     continue
-                    # raise Exception('File "%s" is not valid' % dep)
-                sbuf = os.stat(dep)
-                # TODO: Check/limit size
-                fd = open(dep, 'rb')
-                data = fd.read()
-                fd.close()
                 self.files.append({'name':dep, 'stat':sbuf, 'data':data})
                 depend_ids[dep] = dep
             elif inspect.isfunction(dep) or inspect.isclass(dep) or hasattr(dep, '__class__'):
@@ -440,8 +437,8 @@ class _JobReply(object):
         self.stdout = None
         self.stderr = None
         self.exception = None
-        self.start_time = None
-        self.end_time = None
+        self.start_time = 0
+        self.end_time = 0
 
 class _Cluster(object):
     """Internal use only.
@@ -450,7 +447,7 @@ class _Cluster(object):
 
     def __init__(self, ip_addr=None, ext_ip_addr=None, port=None, node_port=None,
                  shared=False, secret='', keyfile=None, certfile=None):
-        if not hasattr(self, 'ip_addr'):
+        if not hasattr(self, 'asyncoro'):
             self.asyncoro = AsynCoro()
             self.ip_addrs = set()
             if ip_addr:
@@ -512,7 +509,7 @@ class _Cluster(object):
                 self.udp_coro = None
             else:
                 udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-                udp_sock = AsynCoroSocket(udp_sock)
+                udp_sock = AsyncSocket(udp_sock)
                 udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                 udp_sock.bind(('', self.port))
                 self.udp_coro = Coro(self.udp_server, udp_sock)
@@ -531,7 +528,7 @@ class _Cluster(object):
     def send_ping_cluster(self, cluster, coro=None):
         # generator
         ping_msg = {'version':_dispy_version, 'sign':self.sign, 'port':self.port}
-        udp_sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
+        udp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
         udp_sock.settimeout(2)
         for node_spec, node_info in cluster._compute.node_spec.items():
             # TODO: we assume subnets are indicated by '*', instead of
@@ -545,7 +542,7 @@ class _Cluster(object):
                     ping_msg['ip_addr'] = ext_ip_addr
                     bc_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                     bc_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
-                    bc_sock = AsynCoroSocket(bc_sock)
+                    bc_sock = AsyncSocket(bc_sock)
                     bc_sock.settimeout(2)
                     try:
                         yield bc_sock.sendto(b'PING:' + serialize(ping_msg), ('<broadcast>', port))
@@ -566,8 +563,8 @@ class _Cluster(object):
                     except:
                         # logger.debug(traceback.format_exc())
                         pass
-                    tcp_sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                              keyfile=self.keyfile, certfile=self.certfile)
+                    tcp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                                           keyfile=self.keyfile, certfile=self.certfile)
                     tcp_sock.settimeout(2)
                     try:
                         yield tcp_sock.connect((ip_addr, port))
@@ -644,8 +641,8 @@ class _Cluster(object):
         #     port = node_info['port']
         #     if not port:
         #         port = self.node_port
-        #     sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-        #                           keyfile=self.keyfile, certfile=self.certfile)
+        #     sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+        #                        keyfile=self.keyfile, certfile=self.certfile)
         #     sock.settimeout(1)
         #     try:
         #         yield sock.connect((ip_addr, port))
@@ -658,8 +655,8 @@ class _Cluster(object):
         for node in list(cluster._compute.nodes.values()):
             if not node.busy:
                 continue
-            sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                  keyfile=self.keyfile, certfile=self.certfile)
+            sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                               keyfile=self.keyfile, certfile=self.certfile)
             sock.settimeout(1)
             try:
                 yield sock.connect((node.ip_addr, node.port))
@@ -673,8 +670,8 @@ class _Cluster(object):
                         _job = self._sched_jobs.get(uid, None)
                         if _job is None:
                             continue
-                        sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                              keyfile=self.keyfile, certfile=self.certfile)
+                        sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                                           keyfile=self.keyfile, certfile=self.certfile)
                         sock.settimeout(1)
                         yield sock.connect((node.ip_addr, node.port))
                         req = b'RETRIEVE_JOB:' + serialize({'uid':uid, 'hash':_job.hash,
@@ -742,7 +739,7 @@ class _Cluster(object):
 
         if self.shared:
             sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            sock = AsynCoroSocket(sock, keyfile=cluster.keyfile, certfile=cluster.certfile)
+            sock = AsyncSocket(sock, keyfile=cluster.keyfile, certfile=cluster.certfile)
             sock.settimeout(2)
             yield sock.connect((cluster.scheduler_ip_addr, cluster.scheduler_port))
             req = b'DEL_COMPUTE:' + serialize({'ID':cluster._compute.id})
@@ -908,7 +905,7 @@ class _Cluster(object):
                     # logger.debug(traceback.format_exc())
                     continue
                 def _send_pulse(self, pulse_msg, addr, coro=None):
-                    sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
+                    sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
                     sock.settimeout(2)
                     try:
                         yield sock.sendto(b'PULSE:' + serialize(pulse_msg), addr)
@@ -934,7 +931,7 @@ class _Cluster(object):
                     if node.auth_code == info['auth_code']:
                         continue
                 def _get_pong(self, info, coro=None):
-                    sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
+                    sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
                     sock.settimeout(2)
                     try:
                         pong_msg = {'ip_addr':info['scheduler_ip_addr'], 'port':self.port,
@@ -960,7 +957,7 @@ class _Cluster(object):
         # generator
         coro.set_daemon()
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, keyfile=self.keyfile, certfile=self.certfile)
+        sock = AsyncSocket(sock, keyfile=self.keyfile, certfile=self.certfile)
         sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
         if not ip_addr:
             ip_addr = ''
@@ -1023,7 +1020,7 @@ class _Cluster(object):
                 if node:
                     if node.auth_code == info['auth_code']:
                         raise StopIteration
-                sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
+                sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
                 sock.settimeout(2)
                 try:
                     pong_msg = {'ip_addr':info['scheduler_ip_addr'], 'port':self.port,
@@ -1127,13 +1124,12 @@ class _Cluster(object):
                 job.status = reply.status
                 if job.status == DispyJob.Finished:
                     node.jobs += 1
-                if job.end_time is not None:
-                    node.cpu_time += job.end_time - job.start_time
+                node.cpu_time += reply.end_time - reply.start_time
             else:
                 if reply.status == DispyJob.Finished:
                     node.jobs += 1
                 node.busy -= 1
-                node.cpu_time += job.end_time - job.start_time
+                node.cpu_time += reply.end_time - reply.start_time
             yield self.finish_job(_job, reply.status, cluster)
             self._sched_event.set()
 
@@ -1178,7 +1174,7 @@ class _Cluster(object):
                         msg = {'client_ip_addr':cluster._compute.scheduler_ip_addr,
                                'client_port':cluster._compute.scheduler_port}
                         sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
-                        sock = AsynCoroSocket(sock)
+                        sock = AsyncSocket(sock)
                         sock.settimeout(1)
                         yield sock.sendto(b'PULSE:' + serialize(msg),
                                           (cluster.scheduler_ip_addr, cluster.job_result_port))
@@ -1638,7 +1634,7 @@ class JobCluster(object):
                 if dep in depend_ids:
                     continue
                 if compute.type == _Compute.prog_type and not os.path.isfile(dep):
-                    for p in os.environ['PATH'].split(':'):
+                    for p in os.environ['PATH'].split(os.pathsep):
                         f = os.path.join(p, dep)
                         if os.path.isfile(f):
                             logger.debug('Assuming "%s" is program "%s"', dep, f)
@@ -1740,6 +1736,7 @@ class JobCluster(object):
 
     def __exit__(self, exc_type, exc_value, trace):
         self.close()
+        return True
 
     def stats(self):
         """Show statistics for cluster(s).
@@ -1818,7 +1815,7 @@ class SharedJobCluster(JobCluster):
                 raise Exception('"nodes" must be list of IP addresses or host names')
 
         JobCluster.__init__(self, computation, nodes='dummy', depends=depends,
-                            callback=callback, ip_addr=ip_addr, port=0, ext_ip_addr=ext_ip_addr,
+                            callback=callback, ip_addr=ip_addr, port=port, ext_ip_addr=ext_ip_addr,
                             loglevel=loglevel, cleanup=cleanup, pulse_interval=None,
                             poll_interval=None, reentrant=reentrant, fault_recover=fault_recover)
         if pulse_interval is not None:
@@ -1840,6 +1837,10 @@ class SharedJobCluster(JobCluster):
         if not scheduler_port:
             scheduler_port = 51349
 
+        # wait until tcp server has started
+        while self._cluster.port == 0:
+            time.sleep(0.1)
+
         ext_ip_addr = None
         for ext_ip_addr in (self._cluster.ext_ip_addrs - self._cluster.ip_addrs):
             if not ext_ip_addr:
@@ -1850,7 +1851,7 @@ class SharedJobCluster(JobCluster):
                     break
                 
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, blocking=True)
+        sock = AsyncSocket(sock, blocking=True)
         sock.connect((self.scheduler_ip_addr, scheduler_port))
         sock.sendall(self._cluster.auth_code)
         sock.send_msg(b'CLIENT:' + serialize({'version':_dispy_version, 'ip_addr':ext_ip_addr,
@@ -1866,8 +1867,8 @@ class SharedJobCluster(JobCluster):
         self.auth_code = bytes(hashlib.sha1(bytes(reply['sign'] + secret,
                                                   'ascii')).hexdigest(), 'ascii')
 
-        sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM), blocking=True,
-                              keyfile=self.keyfile, certfile=self.certfile)
+        sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM), blocking=True,
+                           keyfile=self.keyfile, certfile=self.certfile)
         sock.settimeout(3)
         self._compute.scheduler_ip_addr = ext_ip_addr
         self._compute.scheduler_port = self._cluster.port
@@ -1894,7 +1895,7 @@ class SharedJobCluster(JobCluster):
             xf.compute_id = self._compute.id
             logger.debug('Sending file "%s"', xf.name)
             sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-            sock = AsynCoroSocket(sock, blocking=True, keyfile=self.keyfile, certfile=self.certfile)
+            sock = AsyncSocket(sock, blocking=True, keyfile=self.keyfile, certfile=self.certfile)
             sock.settimeout(5)
             try:
                 sock.connect((self.scheduler_ip_addr, self.scheduler_port))
@@ -1915,8 +1916,8 @@ class SharedJobCluster(JobCluster):
                 # TODO: delete computation?
             sock.close()
 
-        sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM), blocking=True,
-                              keyfile=self.keyfile, certfile=self.certfile)
+        sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM), blocking=True,
+                           keyfile=self.keyfile, certfile=self.certfile)
         sock.settimeout(3)
         sock.connect((self.scheduler_ip_addr, self.scheduler_port))
         req = b'ADD_COMPUTE:' + serialize({'ID':self._compute.id})
@@ -1950,8 +1951,8 @@ class SharedJobCluster(JobCluster):
             return None
 
         def _submit_job(self, _job, coro=None):
-            sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                  keyfile=self.keyfile, certfile=self.certfile)
+            sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                               keyfile=self.keyfile, certfile=self.certfile)
             sock.settimeout(5)
             try:
                 yield sock.connect((self.scheduler_ip_addr, self.scheduler_port))
@@ -1993,8 +1994,8 @@ class SharedJobCluster(JobCluster):
 
             job.status = DispyJob.Cancelled
             assert self._pending_jobs >= 1
-            sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                  keyfile=self.keyfile, certfile=self.certfile)
+            sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                               keyfile=self.keyfile, certfile=self.certfile)
             sock.settimeout(5)
             try:
                 yield sock.connect((self.scheduler_ip_addr, self.scheduler_port))
@@ -2067,7 +2068,7 @@ def fault_recover_jobs(fault_recover_file, port=51348,
         if job_info['ip_addr'] in node_infos:
             continue
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, blocking=True, keyfile=keyfile, certfile=certfile)
+        sock = AsyncSocket(sock, blocking=True, keyfile=keyfile, certfile=certfile)
         sock.settimeout(2)
         try:
             sock.connect((job_info['ip_addr'], port))
@@ -2091,7 +2092,7 @@ def fault_recover_jobs(fault_recover_file, port=51348,
         if node_info is None:
             continue
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, blocking=True, keyfile=keyfile, certfile=certfile)
+        sock = AsyncSocket(sock, blocking=True, keyfile=keyfile, certfile=certfile)
         sock.settimeout(2)
         try:
             sock.connect((job_info['ip_addr'], node_info['port']))
