@@ -34,7 +34,7 @@ from dispy import _Compute, DispyJob, _DispyJob_, _Node, _JobReply, \
      num_min, _parse_nodes, _node_ipaddr, _XferFile, _dispy_version
 
 import asyncoro
-from asyncoro import Coro, AsynCoro, AsynCoroSocket, MetaSingleton, serialize, unserialize
+from asyncoro import Coro, AsynCoro, AsyncSocket, MetaSingleton, serialize, unserialize
 
 from dispynode import _same_file
 
@@ -67,6 +67,7 @@ class _Cluster(object):
         self.client_ip_addr = None
         self.client_port = None
         self.client_job_result_port = None
+        self.ip_addr = None
         self.pending_results = 0
 
 class _Scheduler(object):
@@ -83,35 +84,28 @@ class _Scheduler(object):
                  cluster_secret='', cluster_keyfile=None, cluster_certfile=None,
                  dest_path_prefix=None, max_file_size=None, zombie_interval=60):
         if not hasattr(self, 'ip_addr'):
-            self.ip_addrs = {}
-            if not ip_addr:
-                ip_addr = socket.gethostbyname(socket.gethostname())
-                # logger.debug('dispy client IP address: %s' % ip_addr)
-            if not isinstance(ip_addr, list):
-                ip_addr = [ip_addr]
+            self.ip_addrs = set()
+            if ip_addr:
+                if not isinstance(ip_addr, list):
+                    ip_addr = [ip_addr]
+                for node in ip_addr:
+                    addr = _node_ipaddr(node)
+                    if addr:
+                        self.ip_addrs.add(addr)
+                    else:
+                        logger.warning('ignoring invalid ip_addr "%s"' % node)
+            if not self.ip_addrs:
+                self.ip_addrs.add(None)
+            self.ext_ip_addrs = set(self.ip_addrs)
             if ext_ip_addr:
                 if not isinstance(ext_ip_addr, list):
                     ext_ip_addr = [ext_ip_addr]
-            else:
-                ext_ip_addr = []
-            for i in range(len(ext_ip_addr), len(ip_addr)):
-                ext_ip_addr.append(None)
-            for ia, ea in zip(ip_addr, ext_ip_addr):
-                addr = _node_ipaddr(ia)
-                if not addr:
-                    logger.warning('ignoring invalid ip_addr "%s"' % addr)
-                    continue
-                if addr.startswith('127.'):
-                    logger.warning('dispy scheduler address "%s" seems to be loopback '
-                                   'address; this will prevent communication with '
-                                   'other nodes.' % ia)
-                ia = addr
-                if ea:
-                    addr = _node_ipaddr(ea)
-                    if not addr:
-                        logger.warning('ignoring invalid ext_ip_addr "%s"' % ea)
-                    ea = addr
-                self.ip_addrs[ia] = ea
+                for node in ext_ip_addr:
+                    addr = _node_ipaddr(node)
+                    if addr:
+                        self.ext_ip_addrs.add(addr)
+                    else:
+                        logger.warning('ignoring invalid ext_ip_addr "%s"' % node)
             if not port:
                 port = 51347
             if not node_port:
@@ -199,14 +193,14 @@ class _Scheduler(object):
 
             self.tcp_coros = []
             self.scheduler_coros = []
-            for ip_addr in self.ip_addrs:
+            for ip_addr in list(self.ip_addrs):
                 self.tcp_coros.append(Coro(self.tcp_server, ip_addr))
                 self.scheduler_coros.append(Coro(self.scheduler_server, ip_addr))
 
             self.udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
             self.udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
             self.udp_sock.bind(('', self.port))
-            self.udp_sock = AsynCoroSocket(self.udp_sock)
+            self.udp_sock = AsyncSocket(self.udp_sock)
             self.udp_coro = Coro(self.udp_server)
 
             self.scheduler_coro = Coro(self._schedule_jobs)
@@ -214,9 +208,18 @@ class _Scheduler(object):
     def scheduler_server(self, ip_addr, coro=None):
         coro.set_daemon()
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, keyfile=self.cluster_keyfile, certfile=self.cluster_certfile)
+        sock = AsyncSocket(sock, keyfile=self.cluster_keyfile, certfile=self.cluster_certfile)
         sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
-        sock.bind((ip_addr, self.scheduler_port))
+        if not ip_addr:
+            ip_addr = ''
+        try:
+            sock.bind((ip_addr, self.scheduler_port))
+        except:
+            if ip_addr == '':
+                ip_addr = None
+            self.ip_addrs.discard(ip_addr)
+            raise StopIteration
+        logger.debug('scheduler at %s:%s' % (ip_addr, self.scheduler_port))
         sock.listen(32)
         while True:
             conn, addr = yield sock.accept()
@@ -229,9 +232,6 @@ class _Scheduler(object):
 
         node_spec = _parse_nodes(self.node_spec)
         ping_msg = {'version':_dispy_version, 'sign':self.sign, 'port':self.port}
-        ext_ip_addrs = set(self.ip_addrs.values())
-        ext_ip_addrs.add(None)
-        ext_ip_addrs = sorted(ext_ip_addrs)
         for node_spec, node_info in node_spec.items():
             # TODO: we assume subnets are indicated by '*', instead of
             # subnet mask; this is a limitation, but specifying with
@@ -240,11 +240,11 @@ class _Scheduler(object):
                 port = node_info['port']
                 if not port:
                     port = self.node_port
-                for ext_ip_addr in ext_ip_addrs:
+                for ext_ip_addr in self.ext_ip_addrs:
                     ping_msg['ip_addr'] = ext_ip_addr
                     bc_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                     bc_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
-                    bc_sock = AsynCoroSocket(bc_sock)
+                    bc_sock = AsyncSocket(bc_sock)
                     bc_sock.settimeout(2)
                     try:
                         yield bc_sock.sendto(b'PING:' + serialize(ping_msg), ('<broadcast>', port))
@@ -258,9 +258,9 @@ class _Scheduler(object):
                 if not port:
                     port = self.node_port
                 ip_addr = node_info['ip_addr']
-                for ext_ip_addr in ext_ip_addrs:
+                for ext_ip_addr in self.ext_ip_addrs:
                     ping_msg['ip_addr'] = ext_ip_addr
-                    udp_sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
+                    udp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
                     udp_sock.settimeout(2)
                     try:
                         yield udp_sock.sendto(b'PING:' + serialize(ping_msg), (ip_addr, port))
@@ -268,8 +268,8 @@ class _Scheduler(object):
                         # logger.debug(traceback.format_exc())
                         pass
                     udp_sock.close()
-                    tcp_sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                              keyfile=self.node_keyfile, certfile=self.node_certfile)
+                    tcp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                                           keyfile=self.node_keyfile, certfile=self.node_certfile)
                     tcp_sock.settimeout(2)
                     try:
                         yield tcp_sock.connect((ip_addr, port))
@@ -301,7 +301,7 @@ class _Scheduler(object):
                         node.last_pulse = time.time()
                         pulse_msg = {'ip_addr':info['scheduler_ip_addr'], 'port':self.port}
                         def _send_pulse(self, pulse_msg, addr, coro=None):
-                            sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
+                            sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
                             sock.settimeout(2)
                             try:
                                 yield sock.sendto(b'PULSE:' + serialize(pulse_msg), addr)
@@ -327,7 +327,7 @@ class _Scheduler(object):
                     if node.auth_code == info['auth_code']:
                         continue
                 def _get_pong(self, info, coro=None):
-                    sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
+                    sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
                     sock.settimeout(2)
                     try:
                         pong_msg = {'ip_addr':info['scheduler_ip_addr'], 'port':self.port,
@@ -352,11 +352,8 @@ class _Scheduler(object):
     def send_ping_cluster(self, cluster, coro=None):
         # generator
         ping_msg = {'version':_dispy_version, 'sign':self.sign, 'port':self.port}
-        udp_sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
+        udp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
         udp_sock.settimeout(2)
-        ext_ip_addrs = set(self.ip_addrs.values())
-        ext_ip_addrs.add(None)
-        ext_ip_addrs = sorted(ext_ip_addrs)
         for node_spec, node_info in cluster._compute.node_spec.items():
             # TODO: we assume subnets are indicated by '*', instead of
             # subnet mask; this is a limitation, but specifying with
@@ -365,11 +362,11 @@ class _Scheduler(object):
                 port = node_info['port']
                 if not port:
                     port = self.node_port
-                for ext_ip_addr in ext_ip_addrs:
+                for ext_ip_addr in self.ext_ip_addrs:
                     ping_msg['ip_addr'] = ext_ip_addr
                     bc_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                     bc_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
-                    bc_sock = AsynCoroSocket(bc_sock)
+                    bc_sock = AsyncSocket(bc_sock)
                     bc_sock.settimeout(2)
                     try:
                         yield bc_sock.sendto(b'PING:' + serialize(ping_msg), ('<broadcast>', port))
@@ -383,15 +380,15 @@ class _Scheduler(object):
                 if not port:
                     port = self.node_port
                 ip_addr = node_info['ip_addr']
-                for ext_ip_addr in ext_ip_addrs:
+                for ext_ip_addr in self.ext_ip_addrs:
                     ping_msg['ip_addr'] = ext_ip_addr
                     try:
                         yield udp_sock.sendto(b'PING:' + serialize(ping_msg), (ip_addr, port))
                     except:
                         # logger.debug(traceback.format_exc())
                         pass
-                    tcp_sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                              keyfile=self.node_keyfile, certfile=self.node_certfile)
+                    tcp_sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                                           keyfile=self.node_keyfile, certfile=self.node_certfile)
                     tcp_sock.settimeout(2)
                     try:
                         yield tcp_sock.connect((ip_addr, port))
@@ -468,8 +465,8 @@ class _Scheduler(object):
         #     port = node_info['port']
         #     if not port:
         #         port = self.node_port
-        #     sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-        #                           keyfile=self.node_keyfile, certfile=self.node_certfile)
+        #     sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+        #                        keyfile=self.node_keyfile, certfile=self.node_certfile)
         #     sock.settimeout(1)
         #     try:
         #         yield sock.connect((ip_addr, port))
@@ -482,8 +479,8 @@ class _Scheduler(object):
         for node in list(cluster._compute.nodes.values()):
             if not node.busy:
                 continue
-            sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                  keyfile=self.node_keyfile, certfile=self.node_certfile)
+            sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                               keyfile=self.node_keyfile, certfile=self.node_certfile)
             sock.settimeout(1)
             try:
                 yield sock.connect((node.ip_addr, node.port))
@@ -497,8 +494,8 @@ class _Scheduler(object):
                         _job = self._sched_jobs.get(uid, None)
                         if _job is None:
                             continue
-                        sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
-                                              keyfile=self.node_keyfile, certfile=self.node_certfile)
+                        sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM),
+                                           keyfile=self.node_keyfile, certfile=self.node_certfile)
                         sock.settimeout(1)
                         yield sock.connect((node.ip_addr, node.port))
                         req = b'RETRIEVE_JOB:' + serialize({'uid':uid, 'hash':_job.hash,
@@ -561,7 +558,7 @@ class _Scheduler(object):
         assert coro is not None
         logger.debug('Sending results for %s to %s, %s', uid, ip, port)
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock)
+        sock = AsyncSocket(sock)
         sock.settimeout(5)
         try:
             yield sock.connect((ip, port))
@@ -658,12 +655,13 @@ class _Scheduler(object):
                 raise StopIteration(None)
             setattr(compute, 'nodes', {})
             cluster = _Cluster(compute)
+            cluster.ip_addr = conn.getsockname()[0]
             compute = cluster._compute
             cluster.client_job_result_port = compute.job_result_port
             cluster.client_ip_addr = compute.scheduler_ip_addr
             cluster.client_port = compute.scheduler_port
             compute.job_result_port = self.port
-            # compute.scheduler_ip_addr = self.ext_ip_addr
+            # compute.scheduler_ip_addr = None
             compute.scheduler_port = self.port
             compute.scheduler_auth = self.auth_code
             compute.id = cluster.id = self.cluster_id
@@ -835,10 +833,9 @@ class _Scheduler(object):
                     if _job.uid == job.uid:
                         del cluster._jobs[i]
                         self.unsched_jobs -= 1
-                        reply = _JobReply(_job, self.ext_ip_addr, status=DispyJob.Cancelled)
+                        reply = _JobReply(_job, cluster.ip_addr, status=DispyJob.Cancelled)
                         Coro(self.send_job_result, _job.uid, compute.id,
-                             cluster.client_ip_addr, cluster.client_job_result_port,
-                             reply)
+                             cluster.client_ip_addr, cluster.client_job_result_port, reply)
                         break
                 else:
                     logger.debug('Invalid job %s!', job.uid)
@@ -973,10 +970,18 @@ class _Scheduler(object):
         # generator
         coro.set_daemon()
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        sock = AsynCoroSocket(sock, keyfile=self.node_keyfile, certfile=self.node_certfile)
+        sock = AsyncSocket(sock, keyfile=self.node_keyfile, certfile=self.node_certfile)
         sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
+        if not ip_addr:
+            ip_addr = ''
+        try:
+            sock.bind((ip_addr, self.port))
+        except:
+            if ip_addr == '':
+                ip_addr = None
+            self.ip_addrs.discard(ip_addr)
+            raise StopIteration
         logger.debug('dispy client at %s:%s' % (ip_addr, self.port))
-        sock.bind((ip_addr, self.port))
         sock.listen(32)
 
         while True:
@@ -1026,7 +1031,7 @@ class _Scheduler(object):
                 if node:
                     if node.auth_code == info['auth_code']:
                         raise StopIteration
-                sock = AsynCoroSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
+                sock = AsyncSocket(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
                 sock.settimeout(2)
                 try:
                     pong_msg = {'ip_addr':info['scheduler_ip_addr'], 'port':self.port,
@@ -1243,14 +1248,14 @@ class _Scheduler(object):
 
         logger.debug('scheduler quitting (%s / %s)', len(self._sched_jobs), self.unsched_jobs)
         for uid, _job in self._sched_jobs.items():
-            reply = _JobReply(_job, self.ext_ip_addr, status=DispyJob.Terminated)
             cluster = self._clusters[_job.compute_id]
+            reply = _JobReply(_job, cluster.ip_addr, status=DispyJob.Terminated)
             compute = cluster._compute
             Coro(self.send_job_result, _job.uid, compute.id,
                  cluster.client_ip_addr, cluster.client_job_result_port, reply)
         for cid, cluster in self._clusters.items():
             for _job in cluster._jobs:
-                reply = _JobReply(_job, self.ext_ip_addr, status=DispyJob.Terminated)
+                reply = _JobReply(_job, cluster.ip_addr, status=DispyJob.Terminated)
                 Coro(self.send_job_result, _job.uid, cluster._compute.id, cluster.client_ip_addr,
                      cluster.client_job_result_port, reply)
             cluster._jobs = []
